use anyhow::bail;

use crate::lerror;
use crate::parser::ast::TokenType;
use crate::parser::expr::parse_expr;
use crate::parser::{Delimiter, Ident, Keyword, Punctuation};
use crate::tokeniser::Token;
use super::ast::typ::Type;
use super::expr::parse_block;
use super::typ::parse_type;
use super::utils;
use super::ast::statement::Statement;

type Result<T> = anyhow::Result<T>;

pub fn parse_statement(tokens: &mut Vec<Token>) -> Result<Option<Statement>> {
    if let Some(_) = utils::check(tokens, TokenType::Keyword(Keyword::Fn)) {
        Ok(Some(parse_fn(tokens)?))
    } else
    if let Some(_) = utils::check(tokens, TokenType::Keyword(Keyword::Type)) {
        Ok(Some(parse_type_alias(tokens)?))
    } else
    if let Some(_) = utils::check(tokens, TokenType::Keyword(Keyword::Const)) {
        Ok(Some(parse_constant(tokens)?))
    } else
    if let Some(_) = utils::check(tokens, TokenType::Keyword(Keyword::Static)) {
        Ok(Some(parse_static(tokens)?))
    } else
    if let Some(_) = utils::check(tokens, TokenType::Keyword(Keyword::Struct)) {
        Ok(Some(parse_struct(tokens)?))
    } else
    if let Some(_) = utils::check(tokens, TokenType::Keyword(Keyword::Enum)) {
        Ok(Some(parse_enum(tokens)?))
    } else
    if let Some(_) = utils::check(tokens, TokenType::Keyword(Keyword::Let)) {
        Ok(Some(parse_let(tokens)?))
    } else {
        Ok(None)
    }
}

fn parse_enum(tokens: &mut Vec<Token>) -> Result<Statement> {
    _ = utils::check_consume(tokens, TokenType::Keyword(Keyword::Enum));
    let name = utils::check_consume_or_err(tokens, TokenType::ident(""), "")?.tt().unwrap_ident();
    _ = utils::check_consume(tokens, TokenType::Delim(Delimiter::CurlyL));
    let mut fields = Vec::new();
    while !tokens.is_empty() {
        if let Some(_) = utils::check_consume(tokens, TokenType::Delim(Delimiter::CurlyR)) {
            break;
        }
        let field_name = utils::check_consume_or_err(tokens, TokenType::ident(""), "")?;
        let loc = field_name.loc().clone();
        let field_name = field_name.tt().unwrap_ident();
        if let None = utils::check_consume(tokens, TokenType::Punct(Punctuation::Comma)) {
            if let None = utils::check(tokens, TokenType::Delim(Delimiter::CurlyR)) {
                lerror!(&loc, "Expected comma after struct field");
                bail!("")
            }
        }
        fields.push(field_name);
    }

    Ok(Statement::Enum { name, fields })
}

fn parse_struct(tokens: &mut Vec<Token>) -> Result<Statement> {
    _ = utils::check_consume(tokens, TokenType::Keyword(Keyword::Struct));
    let name = utils::check_consume_or_err(tokens, TokenType::ident(""), "")?.tt().unwrap_ident();
    _ = utils::check_consume(tokens, TokenType::Delim(Delimiter::CurlyL));
    let mut fields = Vec::new();
    while !tokens.is_empty() {
        if let Some(_) = utils::check_consume(tokens, TokenType::Delim(Delimiter::CurlyR)) {
            break;
        }
        let field_name = utils::check_consume_or_err(tokens, TokenType::ident(""), "")?;
        let loc = field_name.loc().clone();
        let field_name = field_name.tt().unwrap_ident();
        _ = utils::check_consume_or_err(tokens, TokenType::Punct(Punctuation::Colon), "")?;
        let typ = parse_type(tokens)?;
        if let None = utils::check_consume(tokens, TokenType::Punct(Punctuation::Comma)) {
            if let None = utils::check(tokens, TokenType::Delim(Delimiter::CurlyR)) {
                lerror!(&loc, "Expected comma after struct field");
                bail!("")
            }
        }
        fields.push((field_name, typ));
    }

    Ok(Statement::Struct { name, fields })
}

fn parse_static(tokens: &mut Vec<Token>) -> Result<Statement> {
    _ = utils::check_consume(tokens, TokenType::Keyword(Keyword::Static));

    let name = utils::check_consume_or_err(tokens, TokenType::ident(""), "")?.tt().unwrap_ident();
    _ = utils::check_consume_or_err(tokens, TokenType::Punct(Punctuation::Colon), "")?;
    let typ = parse_type(tokens)?;
    let eq = utils::check_consume_or_err(tokens, TokenType::Punct(Punctuation::Eq), "")?;
    let Some(val) = parse_expr(tokens, 0, false)? else {
        lerror!(eq.loc(), "Expected expression found nothing");
        bail!("")
    };
    _ = utils::check_consume_or_err(tokens, TokenType::Punct(Punctuation::Semi), "")?;
    Ok(Statement::StaticVar { name, typ, val })
}

fn parse_let(tokens: &mut Vec<Token>) -> Result<Statement> {
    _ = utils::check_consume(tokens, TokenType::Keyword(Keyword::Let));
    let name = utils::check_consume_or_err(tokens, TokenType::ident(""), "")?.tt().unwrap_ident();
    let mut typ = None;
    let mut val = None;
    if let Some(_) = utils::check_consume(tokens, TokenType::Punct(Punctuation::Colon)) {
        typ = Some(parse_type(tokens)?);
    }
    if let Some(eq) = utils::check_consume(tokens, TokenType::Punct(Punctuation::Eq)) {
        let Some(_val) = parse_expr(tokens, 0, false)? else {
            lerror!(eq.loc(), "Expected expression found nothing");
            bail!("")
        };
        val = Some(_val);
    }
    _ = utils::check_consume_or_err(tokens, TokenType::Punct(Punctuation::Semi), "")?;
    Ok(Statement::Let { name, typ, val })
}
fn parse_constant(tokens: &mut Vec<Token>) -> Result<Statement> {
    _ = utils::check_consume(tokens, TokenType::Keyword(Keyword::Const));

    if let Some(_) = utils::check(tokens, TokenType::Keyword(Keyword::Fn)) {
        unimplemented!()
    }
    let name = utils::check_consume_or_err(tokens, TokenType::ident(""), "")?.tt().unwrap_ident();
    _ = utils::check_consume_or_err(tokens, TokenType::Punct(Punctuation::Colon), "")?;
    let typ = parse_type(tokens)?;
    let eq = utils::check_consume_or_err(tokens, TokenType::Punct(Punctuation::Eq), "")?;
    let Some(val) = parse_expr(tokens, 0, false)? else {
        lerror!(eq.loc(), "Expected expression found nothing");
        bail!("")
    };
    _ = utils::check_consume_or_err(tokens, TokenType::Punct(Punctuation::Semi), "")?;
    Ok(Statement::ConstVar { name, typ, val })
}

fn parse_type_alias(tokens: &mut Vec<Token>) -> Result<Statement> {
    _ = utils::check_consume(tokens, TokenType::Keyword(Keyword::Type));
    let name = utils::check_consume_or_err(tokens, TokenType::ident(""), "")?.tt().unwrap_ident();
    _ = utils::check_consume_or_err(tokens, TokenType::Punct(Punctuation::Eq), "")?;
    let typ = parse_type(tokens)?;
    _ = utils::check_consume_or_err(tokens, TokenType::Punct(Punctuation::Semi), "")?;

    Ok(Statement::TypeAlias { name, typ })
}

fn parse_fn(tokens: &mut Vec<Token>) -> Result<Statement> {
    // Just remove the kw since we checked it before
    _ = utils::check_consume(tokens, TokenType::Keyword(Keyword::Fn));

    let mut struct_name = None;
    let mut name = utils::check_consume_or_err(tokens, TokenType::ident(""), "")?.tt().unwrap_ident();
    // Check if this is a struct method
    if let Some(_) = utils::check_consume(tokens, TokenType::Punct(Punctuation::Fieldaccess)) {
        struct_name = Some(name);   
        name = utils::check_consume_or_err(tokens, TokenType::ident(""), "")?.tt().unwrap_ident();
    }
    let params = parse_fn_params(tokens)?;

    // Check for return type cause it optional
    let mut ret_type = None;
    if let Some(_) = utils::check_consume(tokens, TokenType::Punct(Punctuation::Arrow)) {
        ret_type = Some(parse_type(tokens)?);
    }
    let body;
    if let Some(_) = utils::check(tokens, TokenType::Delim(Delimiter::CurlyL)) {
        body = Some(parse_block(tokens)?);
    } else {
        // Check if its just a declaration
        _ = utils::check_consume_or_err(tokens, TokenType::Punct(Punctuation::Semi), "")?;
        body = None;
    }
    Ok(Statement::Fn {
        struct_name, 
        name, 
        params, 
        ret_type, 
        qual_const: false, 
        qual_extern: None, 
        body,
    })
}



fn parse_fn_params(tokens: &mut Vec<Token>) -> Result<Vec<(Ident, Type)>> {
    let mut args = Vec::new();
    utils::check_consume_or_err(tokens, TokenType::Delim(Delimiter::ParenL), "")?;
    while !tokens.is_empty() {
        let name = utils::check_consume_or_err(tokens, TokenType::ident(""), "")?;
        utils::check_consume_or_err(tokens, TokenType::Punct(Punctuation::Colon), "")?;    
        //dbg!(&name);
        let typ = parse_type(tokens)?;
        args.push((name.tt().unwrap_ident(), typ));

        if let None = utils::check_consume(tokens, TokenType::Punct(Punctuation::Comma)) {
            break;
        }
    }

    utils::check_consume_or_err(tokens, TokenType::Delim(Delimiter::ParenR), "")?;
    Ok(args)
}


