use std::sync::{Arc, Mutex};

use lazy_static::lazy_static;

use crate::{common::Loc, debug, lerror, tokeniser::Token};

use super::ast::TokenType;


lazy_static!(
    static ref LAST_LOC: Arc<Mutex<Loc>> = Arc::new(Mutex::new(Loc::default()));
);

pub fn check(tokens: &Vec<Token>, tt: TokenType) -> Option<&Token> {
    if let Some(tkn) = tokens.last() {
        if tkn.tt() == &tt ||
            // ignore internal values if searching for these
            matches!((tkn.tt(), &tt), (TokenType::Ident(_), TokenType::Ident(_))) ||
            matches!((tkn.tt(), &tt), (TokenType::String(_), TokenType::String(_))) ||
            matches!((tkn.tt(), &tt), (TokenType::Number(_), TokenType::Number(_))) ||
            matches!((tkn.tt(), &tt), (TokenType::Char(_), TokenType::Char(_))) 
        {
            debug!("check: {}", tkn);
            return Some(tkn);
        }
    }
    None
}
pub fn check_2_last(tokens: &Vec<Token>, tt: TokenType) -> Option<&Token> {
    if tokens.len() < 2 {
        return None
    }
    if let Some(tkn) = tokens.get(tokens.len() - 2) {
        if tkn.tt() == &tt ||
            // ignore internal values if searching for these
            matches!((tkn.tt(), &tt), (TokenType::Ident(_), TokenType::Ident(_))) ||
            matches!((tkn.tt(), &tt), (TokenType::String(_), TokenType::String(_))) ||
            matches!((tkn.tt(), &tt), (TokenType::Number(_), TokenType::Number(_))) ||
            matches!((tkn.tt(), &tt), (TokenType::Char(_), TokenType::Char(_))) 
        {
            return Some(tkn);
        }
    }
    None
}

pub fn check_consume(tokens: &mut Vec<Token>, tt: TokenType) -> Option<Token> {
    if let Some(tkn) = tokens.last() {
        if tkn.tt() == &tt ||
            // ignore internal values if searching for these
            matches!((tkn.tt(), &tt), (TokenType::Ident(_), TokenType::Ident(_))) ||
            matches!((tkn.tt(), &tt), (TokenType::String(_), TokenType::String(_))) ||
            matches!((tkn.tt(), &tt), (TokenType::Number(_), TokenType::Number(_))) ||
            matches!((tkn.tt(), &tt), (TokenType::Char(_), TokenType::Char(_))) 
        {
            *LAST_LOC.lock().expect("Could not lock LAST_LOC") = tkn.loc().clone();
            debug!("check_consume: {}", tokens.last()?);
            return Some(tokens.pop()?);
        }
    }
    None
}

pub fn check_consume_or_err(tokens: &mut Vec<Token>, tt: TokenType, err_msg: &'static str) -> anyhow::Result<Token> {
    if let Some(tkn) = tokens.last() {
        if tkn.tt() == &tt ||
            // ignore internal values if searching for these
            matches!((tkn.tt(), &tt), (TokenType::Ident(_), TokenType::Ident(_))) ||
            matches!((tkn.tt(), &tt), (TokenType::String(_), TokenType::String(_))) ||
            matches!((tkn.tt(), &tt), (TokenType::Number(_), TokenType::Number(_))) ||
            matches!((tkn.tt(), &tt), (TokenType::Char(_), TokenType::Char(_))) 
        {
            *LAST_LOC.lock().expect("Could not lock LAST_LOC") = tkn.loc().clone();
            return Ok(tokens.pop().expect("Unreachable"));
        } else {
            lerror!(tkn.loc(), "Expected: '{tt}', got: '{}': {err_msg}", tkn.tt());
            //anyhow::bail!(format!("{}: ERROR: Expected: '{tt:?}', got: '{:?}': {err_msg}", tkn.loc(), tkn.tt()))
            anyhow::bail!("")
        }
    }
    let loc = LAST_LOC.lock().expect("Could not lock LAST_LOC");
    lerror!(&loc, "Expected: '{tt}', got: '(empty)': {err_msg}");
    // anyhow::bail!(format!("{loc}: ERROR: Expected '{tt:?}', got (empty): {err_msg}"))
    anyhow::bail!("")
}

pub fn check_consume_from_many(tokens: &mut Vec<Token>, tts: &[TokenType]) -> Option<Token> {
    for tt in tts {
        if let Some(tkn) = check_consume(tokens, tt.clone()) {
            return Some(tkn);
        }
    }
    None
}

pub fn check_from_many<'a>(tokens: &'a mut Vec<Token>, tts: &[TokenType]) -> Option<&'a Token> {
    for tt in tts {
        if let Some(tkn) = check(tokens, tt.clone()) {
            return Some(tkn);
        }
    }
    None
}

pub fn check_consume_or_err_from_many(tokens: &mut Vec<Token>, tts: &[TokenType], err_msg: &'static str) -> anyhow::Result<Token> {
    if let Some(tkn) = tokens.last() {
        for tt in tts {
            if tkn.tt() == tt ||
                // ignore internal values if searching for these
                matches!((tkn.tt(), &tt), (TokenType::Ident(_), TokenType::Ident(_))) ||
                matches!((tkn.tt(), &tt), (TokenType::String(_), TokenType::String(_))) ||
                matches!((tkn.tt(), &tt), (TokenType::Number(_), TokenType::Number(_))) ||
                matches!((tkn.tt(), &tt), (TokenType::Char(_), TokenType::Char(_))) 
            {
                *LAST_LOC.lock().expect("Could not lock LAST_LOC") = tkn.loc().clone();
                return Ok(tokens.pop().expect("Unreachable"));
            } 
        }
        lerror!(tkn.loc(), "Expected: '{tts:?}', got: '{}': {err_msg}", tkn.tt());
        anyhow::bail!("")
    }
    let loc = LAST_LOC.lock().expect("Could not lock LAST_LOC");
    lerror!(&loc, "Expected: '{tts:?}', got: '(empty)': {err_msg}");
    anyhow::bail!("")
}

pub fn get_last_loc() -> Loc {
    LAST_LOC.lock().expect("Could not lock LAST_LOC").clone()
}
